#!/usr/bin/python
# Python script to bootstrap the creation of a new library within MRPT
#  Invoke without arguments from the MRPT root directory, that is:
#   $ scripts/bootstrap_new_lib.py
#
#  or from the MRPT/scripts directory:
#   $ ./bootstrap_new_lib.py
#
#  Script by Jose Luis Blanco, July 2010.

import os
import sys
import re
import string

#------   MAIN   -------
def main():
	if os.path.exists(os.path.normpath("../version_prefix.txt")):
		os.chdir('..')

	if not os.path.exists(os.path.normpath("./version_prefix.txt")):
		print("ERROR: Invoke this script from the <MRPT> ROOT directory")
		return 1  # end

	# Make sure our CWD is the root of MRPT source tree:
	assert os.path.exists("libs")

	print("Enter the name of the new library (example, for 'mrpt-foo', enter 'foo'):")
	NewLibName = input(">")

	# Replace string lists:
	sFrom = ['@NAME@', '@name@']
	sTo = [NewLibName.upper(), NewLibName.lower()]

	parseDir = os.path.normpath("parse-files/new_lib_bootstrap")
	baseDir  = os.path.normpath("libs/" + NewLibName )

	assert os.path.exists(parseDir)
	assert not os.path.exists(baseDir)

	# Create new lib:
	# --------------------------------
	print("Creating new lib in directory: " + baseDir)

	os.mkdir(baseDir)   # Create dirs
	os.mkdir(os.path.normpath(baseDir+"/include"))
	os.mkdir(os.path.normpath(baseDir+"/include/mrpt"))
	os.mkdir(os.path.normpath(baseDir+"/include/mrpt/"+ NewLibName.lower() ))
	os.mkdir(os.path.normpath(baseDir+"/src"))

	replaceInFile(
		os.path.normpath(parseDir + '/CMakeLists.txt.in'),
		os.path.normpath(baseDir + '/CMakeLists.txt'),
		sFrom, sTo)

	replaceInFile(
		os.path.normpath(parseDir+'/include/mrpt/name.h.in'),
		os.path.normpath(baseDir+'/include/mrpt/'+ NewLibName.lower() + '.h'),
		sFrom, sTo)

	replaceInFile(
		os.path.normpath(parseDir+'/src/name-precomp.cpp.in'),
		os.path.normpath(baseDir+'/src/'+ NewLibName.lower() +'-precomp.cpp'),
		sFrom, sTo)
	replaceInFile(
		os.path.normpath(parseDir+'/src/name-precomp.h.in'),
		os.path.normpath(baseDir+'/src/'+ NewLibName.lower() +'-precomp.h'),
		sFrom, sTo)

	replaceInFile(
		os.path.normpath(parseDir+'/src/registerAllClasses.cpp.in'),
		os.path.normpath(baseDir+'/src/registerAllClasses.cpp'),
		sFrom, sTo)


	return 0  # OK
#-----------


def replaceInFile(fileIn, fileOut, textToSearch, newText) :
	o = open(fileOut,"w")
	data = open(fileIn).read()
	for idx in range(len(textToSearch)):
		data = re.sub(textToSearch[idx], newText[idx], data)
	o.write(data)
	o.close()
#-----------
    

if __name__ == "__main__":
	sys.exit(main())



#!/usr/bin/env python

"""
Script decimates a ground truth file. File should be in the MRPT
GridMapNavSimul format. Use the --help flag for more on the available user
options

"""

import argparse

# Argument Parsing
parser = argparse.ArgumentParser(description='Decimate a ground-truth file by the decimation ratio provided')
parser.add_argument('-i', '--input_gt',
                    default=None,
                    help='Specify the input ground-truth file')
parser.add_argument('-o', '--output_gt',
                    default=None,
                    help='Specify the output file for the decimated ground-truth')
parser.add_argument('-d', '--decimation_ratio',
                    default=2,
                    type=int,
                    help='Specify the decimation ratio. Keep one out of d poses, where d is the decimation ratio')

args = vars(parser.parse_args())

fname_in = args["input_gt"]
fname_out = args["output_gt"]
dec_ratio = args["decimation_ratio"]

print("ground-truth *input* file: {}".format(fname_in))
print("ground-truth *output* file: {}".format(fname_out))
print("Decimate and keep one every *{}* poses".format(dec_ratio))

# parse input file
with open(fname_in, 'r') as f_in:
    lines = f_in.readlines()

# populate output file
with open(fname_out, 'w') as f_out:
    for i, line in enumerate(lines):
        if i % dec_ratio == 0:
            f_out.write(str(line))

print("Decimation finished.")



# python/example3.py -- running NanoGUI in detached mode
# (contributed by Dmitriy Morozov)
#
# NanoGUI was developed by Wenzel Jakob <wenzel@inf.ethz.ch>.
# The widget drawing code is based on the NanoVG demo application
# by Mikko Mononen.
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE.txt file.

import nanogui
from nanogui import *
import time


class TestApp(Screen):
    def __init__(self):
        super(TestApp, self).__init__((190, 170), "NanoGUI Test")

        window = Window(self, "Detached mode")
        window.setPosition((15, 15))
        window.setLayout(GroupLayout())

        Label(window, "Push buttons", "sans-bold")
        b = Button(window, "Plain button")

        def cb():
            print("pushed!")
        b.setCallback(cb)

        b = Button(window, "Quit")

        def cb2():
            self.setVisible(False)
        b.setCallback(cb2)

        self.performLayout()


if __name__ == "__main__":
    nanogui.init()
    test = TestApp()
    test.drawAll()
    test.setVisible(True)

    print("Launching detached mainloop")
    h = nanogui.mainloop(detach=test)

    print("Back in Python context")
    for i in range(10):
        print(i)
        time.sleep(1)
        if not nanogui.active():
            break

    h.join()
    nanogui.shutdown()



# python/example2.py -- Python version of an example application that shows
# how to use the form helper class. For a C++ implementation, see
# '../src/example2.cpp'.
#
# NanoGUI was developed by Wenzel Jakob <wenzel@inf.ethz.ch>.
# The widget drawing code is based on the NanoVG demo application
# by Mikko Mononen.
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE.txt file.

import nanogui
import math
import gc

from nanogui import Screen, FormHelper

bvar = True
ivar = 12345678
dvar = math.pi
strvar = "A string"
enumvar = 1
colvar = nanogui.Color(.5, .5, .7, 1)


def make_accessors(name):
    def setter(value):
        globals()[name] = value

    def getter():
        return globals()[name]
    return setter, getter

nanogui.init()

use_gl_4_1 = False # Set to True to create an OpenGL 4.1 context.
if use_gl_4_1:
    # NanoGUI presents many options for you to utilize at your discretion.
    # See include/nanogui/screen.h for what all of the options are.
    screen = Screen((500, 700), "NanoGUI test [GL 4.1]", glMajor=4, glMinor=1)
else:
    screen = Screen((500, 700), "NanoGUI test")

gui = FormHelper(screen)
window = gui.addWindow((10, 10), "Form helper example")

gui.addGroup("Basic types")
gui.addBoolVariable("bool", *make_accessors("bvar"))
gui.addStringVariable("string", *make_accessors("strvar"))

gui.addGroup("Validating fields")
gui.addIntVariable("int", *make_accessors("ivar"))
gui.addDoubleVariable("double", *make_accessors("dvar"))

gui.addGroup("Complex types")
gui.addEnumVariable("Enumeration", *make_accessors("enumvar")) \
   .setItems(["Item 1", "Item 2", "Item 3"])

def cp_final_cb(color):
    print(
        "ColorPicker Final Callback: [{0}, {1}, {2}, {3}]".format(color.r,
                                                                  color.g,
                                                                  color.b,
                                                                  color.w)
    )


gui.addColorVariable("Color", *make_accessors("colvar")).setFinalCallback(cp_final_cb)

gui.addGroup("Other widgets")


def cb():
    print("Button pressed.")
gui.addButton("A button", cb)

screen.setVisible(True)
screen.performLayout()
window.center()

nanogui.mainloop()
screen = gui = window = None
gc.collect()
nanogui.shutdown()



#!/usr/bin/env python3
#
#  This file is a modification of the `mkdoc.py` script in pybind11:
#      - File: https://github.com/pybind/pybind11/blob/master/tools/mkdoc.py
#      - License: https://github.com/pybind/pybind11/blob/master/LICENSE
#
#  Where it differs:
#  - process_comment: added substitutions.  cheap hack: turn all rst into markdown ```
#        - Not necessarily ideal, but at least doesn't break the tables
#
#        s = re.sub(r'\\class\s+%s.*' % cpp_group, '', s)   # removes \class Thing file.h nanogui/file.h
#        s = re.sub(r'\\struct\s+%s.*' % cpp_group, '', s)  # removes \struct Thing file.h nanogui/file.h
#        s = re.sub(r'\\rst', r'```\n\n', s)
#        s = re.sub(r'\\endrst', r'```\n\n', s)
#        s = re.sub(r'.. note::', r'Note:', s)
#        s = re.sub(r'.. warning::', r'Warning:', s)
#        s = re.sub(r'.. danger::', r'Danger:', s)
#        s = re.sub(r'.. code-block::\s*\w*', r'', s)
#
#  - process_comment: near the end, have to treat .. code-block:: specially
#
#        for line in x.splitlines():
#            if len(line) < 4:
#                result += line.strip()
#            else:
#                # this is a .. code-block:: indentation (three spaces)
#                if line.startswith('   ') and line[3] != ' ':
#                    result += line[3:].strip() + '\n'
#                else:
#                    result += line.strip() + '\n'
#
#  Syntax: mkdoc.py [-I<path> ..] [.. a list of header files ..]
#
#  Extract documentation from C++ header files to use it in Python bindings
#

import os
import sys
import platform
import re
import textwrap

# CMake generated target is explicitly calling `python3`, this sanity check is
# to support the fact that some `pip3` installable `clang` libraries are
# packaged for python3, but only actually work in python2.
if sys.version[0] != "3":
    raise RuntimeError("`mkdoc_rst.py` must be run with Python3.")


# If user has `clang` already installed, try and use that.  There are a couple
# of unofficial options for python **3**, two that seem to work are `clang-5`
# and `libclang-py3`.  It likely depends on what you have installed on your
# system, so your mileage will very likely vary.
try:
    from clang import cindex
    from clang.cindex import CursorKind
except:
    # Fallback on pybind11
    # This file is tailored to the NanoGUI documentation build system, the clang
    # module required is present in a full recursive clone here.
    base_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    clang_parent_folder = os.path.join(base_path, "ext/pybind11/tools/")
    if not os.path.isdir(clang_parent_folder):
        raise RuntimeError(
             "The NanoGUI dependencies repository (pybind11, etc.) appear to be missing!\n"
             "You probably did not clone the project with --recursive. It is possible to recover\n"
             "by calling 'git submodule update --init --recursive'"
        )
    else:
        sys.path.insert(0, clang_parent_folder)

    # Now we can import clang
    from clang import cindex
    from clang.cindex import CursorKind

from collections import OrderedDict
from threading import Thread, Semaphore
from multiprocessing import cpu_count

RECURSE_LIST = [
    CursorKind.TRANSLATION_UNIT,
    CursorKind.NAMESPACE,
    CursorKind.CLASS_DECL,
    CursorKind.STRUCT_DECL,
    CursorKind.ENUM_DECL,
    CursorKind.CLASS_TEMPLATE
]

PRINT_LIST = [
    CursorKind.CLASS_DECL,
    CursorKind.STRUCT_DECL,
    CursorKind.ENUM_DECL,
    CursorKind.ENUM_CONSTANT_DECL,
    CursorKind.CLASS_TEMPLATE,
    CursorKind.FUNCTION_DECL,
    CursorKind.FUNCTION_TEMPLATE,
    CursorKind.CONVERSION_FUNCTION,
    CursorKind.CXX_METHOD,
    CursorKind.CONSTRUCTOR,
    CursorKind.FIELD_DECL
]

CPP_OPERATORS = {
    '<=': 'le', '>=': 'ge', '==': 'eq', '!=': 'ne', '[]': 'array',
    '+=': 'iadd', '-=': 'isub', '*=': 'imul', '/=': 'idiv', '%=':
    'imod', '&=': 'iand', '|=': 'ior', '^=': 'ixor', '<<=': 'ilshift',
    '>>=': 'irshift', '++': 'inc', '--': 'dec', '<<': 'lshift', '>>':
    'rshift', '&&': 'land', '||': 'lor', '!': 'lnot', '~': 'bnot',
    '&': 'band', '|': 'bor', '+': 'add', '-': 'sub', '*': 'mul', '/':
    'div', '%': 'mod', '<': 'lt', '>': 'gt', '=': 'assign', '()': 'call'
}

CPP_OPERATORS = OrderedDict(
    sorted(CPP_OPERATORS.items(), key=lambda t: -len(t[0])))

job_count = cpu_count()
job_semaphore = Semaphore(job_count)

registered_names = dict()


def d(s):
    if type(s) is bytes:
        return s.decode('utf8')
    elif type(s) is str:
        return s
    else:
        raise RuntimeError(
            "d(s): `s` cannot be decoded, it's type is: {0}".format(type(s))
        )


def sanitize_name(name):
    global registered_names
    name = re.sub(r'type-parameter-0-([0-9]+)', r'T\1', name)
    for k, v in CPP_OPERATORS.items():
        name = name.replace('operator%s' % k, 'operator_%s' % v)
    name = re.sub('<.*>', '', name)
    name = ''.join([ch if ch.isalnum() else '_' for ch in name])
    name = re.sub('_$', '', re.sub('_+', '_', name))
    if name in registered_names:
        registered_names[name] += 1
        name += '_' + str(registered_names[name])
    else:
        registered_names[name] = 1
    return '__doc_' + name


def process_comment(comment):
    result = ''

    # Remove C++ comment syntax
    leading_spaces = float('inf')
    for s in comment.expandtabs(tabsize=4).splitlines():
        s = s.strip()
        if s.startswith('/*'):
            s = s[2:].lstrip('*')
        elif s.endswith('*/'):
            s = s[:-2].rstrip('*')
        elif s.startswith('///<'):
            s = s[4:]
        elif s.startswith('///'):
            s = s[3:]
        if s.startswith('*'):
            s = s[1:]
        if len(s) > 0:
            leading_spaces = min(leading_spaces, len(s) - len(s.lstrip()))
        result += s + '\n'

    if leading_spaces != float('inf'):
        result2 = ""
        for s in result.splitlines():
            result2 += s[leading_spaces:] + '\n'
        result = result2

    # Doxygen tags
    cpp_group = '([\w:]+)'
    param_group = '([\[\w:\]]+)'

    s = result

    s = re.sub(r'\\class\s+%s.*' % cpp_group, '', s)   # removes \class Thing file.h nanogui/file.h
    s = re.sub(r'\\struct\s+%s.*' % cpp_group, '', s)  # removes \struct Thing file.h nanogui/file.h
    s = re.sub(r'\\rst', r'```\n\n', s)
    s = re.sub(r'\\endrst', r'```\n\n', s)
    s = re.sub(r'.. note::', r'Note:', s)
    s = re.sub(r'.. warning::', r'Warning:', s)
    s = re.sub(r'.. danger::', r'Danger:', s)
    s = re.sub(r'.. code-block::\s*\w*', r'', s)

    s = re.sub(r'\\c\s+%s' % cpp_group, r'``\1``', s)
    s = re.sub(r'\\a\s+%s' % cpp_group, r'*\1*', s)
    s = re.sub(r'\\e\s+%s' % cpp_group, r'*\1*', s)
    s = re.sub(r'\\em\s+%s' % cpp_group, r'*\1*', s)
    s = re.sub(r'\\b\s+%s' % cpp_group, r'**\1**', s)
    s = re.sub(r'\\ingroup\s+%s' % cpp_group, r'', s)
    s = re.sub(r'\\param%s?\s+%s' % (param_group, cpp_group),
               r'\n\n$Parameter ``\2``:\n\n', s)
    s = re.sub(r'\\tparam%s?\s+%s' % (param_group, cpp_group),
               r'\n\n$Template parameter ``\2``:\n\n', s)

    for in_, out_ in {
        'return': 'Returns',
        'author': 'Author',
        'authors': 'Authors',
        'copyright': 'Copyright',
        'date': 'Date',
        'remark': 'Remark',
        'sa': 'See also',
        'see': 'See also',
        'extends': 'Extends',
        'throws': 'Throws',
        'throw': 'Throws'
    }.items():
        s = re.sub(r'\\%s\s*' % in_, r'\n\n$%s:\n\n' % out_, s)

    s = re.sub(r'\\details\s*', r'\n\n', s)
    s = re.sub(r'\\brief\s*', r'', s)
    s = re.sub(r'\\short\s*', r'', s)
    s = re.sub(r'\\ref\s*', r'', s)

    s = re.sub(r'\\code\s?(.*?)\s?\\endcode',
               r"```\n\1\n```\n", s, flags=re.DOTALL)

    # HTML/TeX tags
    s = re.sub(r'<tt>(.*?)</tt>', r'``\1``', s, flags=re.DOTALL)
    s = re.sub(r'<pre>(.*?)</pre>', r"```\n\1\n```\n", s, flags=re.DOTALL)
    s = re.sub(r'<em>(.*?)</em>', r'*\1*', s, flags=re.DOTALL)
    s = re.sub(r'<b>(.*?)</b>', r'**\1**', s, flags=re.DOTALL)
    s = re.sub(r'\\f\$(.*?)\\f\$', r'$\1$', s, flags=re.DOTALL)
    s = re.sub(r'<li>', r'\n\n* ', s)
    s = re.sub(r'</?ul>', r'', s)
    s = re.sub(r'</li>', r'\n\n', s)

    s = s.replace('``true``', '``True``')
    s = s.replace('``false``', '``False``')

    # Re-flow text
    wrapper = textwrap.TextWrapper()
    wrapper.expand_tabs = True
    wrapper.replace_whitespace = True
    wrapper.drop_whitespace = True
    wrapper.width = 70
    wrapper.initial_indent = wrapper.subsequent_indent = ''

    result = ''
    in_code_segment = False
    for x in re.split(r'(```)', s):
        if x == '```':
            if not in_code_segment:
                result += '```\n'
            else:
                result += '\n```\n\n'
            in_code_segment = not in_code_segment
        elif in_code_segment:
            for line in x.splitlines():
                if len(line) < 4:
                    result += line.strip()
                else:
                    # this is a .. code-block:: indentation (three spaces),
                    # strip leading three spaces, preserve remaining indentation
                    if line.startswith('   '):
                        result += line[3:].rstrip() + '\n'
                    else:
                        result += line.strip() + '\n'
        else:
            for y in re.split(r'(?: *\n *){2,}', x):
                wrapped = wrapper.fill(re.sub(r'\s+', ' ', y).strip())
                if len(wrapped) > 0 and wrapped[0] == '$':
                    result += wrapped[1:] + '\n'
                    wrapper.initial_indent = \
                        wrapper.subsequent_indent = ' ' * 4
                else:
                    if len(wrapped) > 0:
                        result += wrapped + '\n\n'
                    wrapper.initial_indent = wrapper.subsequent_indent = ''
    return result.rstrip().lstrip('\n')


def extract(filename, node, prefix, output):
    num_extracted = 0
    if not (node.location.file is None or
            os.path.samefile(d(node.location.file.name), filename)):
        return 0
    if node.kind in RECURSE_LIST:
        sub_prefix = prefix
        if node.kind != CursorKind.TRANSLATION_UNIT:
            if len(sub_prefix) > 0:
                sub_prefix += '_'
            sub_prefix += d(node.spelling)
        for i in node.get_children():
            num_extracted += extract(filename, i, sub_prefix, output)
        if num_extracted == 0:
            return 0
    if node.kind in PRINT_LIST:
        comment = d(node.raw_comment) if node.raw_comment is not None else ''
        comment = process_comment(comment)
        sub_prefix = prefix
        if len(sub_prefix) > 0:
            sub_prefix += '_'
        if len(node.spelling) > 0:
            name = sanitize_name(sub_prefix + d(node.spelling))
            output.append('\nstatic const char *%s =%sR"doc(%s)doc";' %
                (name, '\n' if '\n' in comment else ' ', comment))
            num_extracted += 1
    return num_extracted


class ExtractionThread(Thread):
    def __init__(self, filename, parameters, output):
        Thread.__init__(self)
        self.filename = filename
        self.parameters = parameters
        self.output = output
        job_semaphore.acquire()

    def run(self):
        print('Processing "%s" ..' % self.filename, file=sys.stderr)
        try:
            index = cindex.Index(
                cindex.conf.lib.clang_createIndex(False, True))
            tu = index.parse(self.filename, self.parameters)
            extract(self.filename, tu.cursor, '', self.output)
        finally:
            job_semaphore.release()

if __name__ == '__main__':
    parameters = ['-x', 'c++', '-std=c++11']
    filenames = []

    if platform.system() == 'Darwin':
        dev_path = '/Applications/Xcode.app/Contents/Developer/'
        lib_dir = dev_path + 'Toolchains/XcodeDefault.xctoolchain/usr/lib/'
        sdk_dir = dev_path + 'Platforms/MacOSX.platform/Developer/SDKs'
        libclang = lib_dir + 'libclang.dylib'

        if os.path.exists(libclang):
            cindex.Config.set_library_path(os.path.dirname(libclang))

        if os.path.exists(sdk_dir):
            sysroot_dir = os.path.join(sdk_dir, next(os.walk(sdk_dir))[1][0])
            parameters.append('-isysroot')
            parameters.append(sysroot_dir)

    for item in sys.argv[1:]:
        if item.startswith('-'):
            parameters.append(item)
        else:
            filenames.append(item)

    if len(filenames) == 0:
        print('Syntax: %s [.. a list of header files ..]' % sys.argv[0])
        exit(-1)

    print('''/*
  This file contains docstrings for the Python bindings.
  Do not edit! These were automatically extracted by mkdoc.py
 */

#define __EXPAND(x)                                      x
#define __COUNT(_1, _2, _3, _4, _5, _6, _7, COUNT, ...)  COUNT
#define __VA_SIZE(...)                                   __EXPAND(__COUNT(__VA_ARGS__, 7, 6, 5, 4, 3, 2, 1))
#define __CAT1(a, b)                                     a ## b
#define __CAT2(a, b)                                     __CAT1(a, b)
#define __DOC1(n1)                                       __doc_##n1
#define __DOC2(n1, n2)                                   __doc_##n1##_##n2
#define __DOC3(n1, n2, n3)                               __doc_##n1##_##n2##_##n3
#define __DOC4(n1, n2, n3, n4)                           __doc_##n1##_##n2##_##n3##_##n4
#define __DOC5(n1, n2, n3, n4, n5)                       __doc_##n1##_##n2##_##n3##_##n4##_##n5
#define __DOC6(n1, n2, n3, n4, n5, n6)                   __doc_##n1##_##n2##_##n3##_##n4##_##n5##_##n6
#define __DOC7(n1, n2, n3, n4, n5, n6, n7)               __doc_##n1##_##n2##_##n3##_##n4##_##n5##_##n6##_##n7
#define DOC(...)                                         __EXPAND(__EXPAND(__CAT2(__DOC, __VA_SIZE(__VA_ARGS__)))(__VA_ARGS__))

#if defined(__GNUG__)
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-variable"
#endif
''')

    output = []
    for filename in filenames:
        thr = ExtractionThread(filename, parameters, output)
        thr.start()

    print('Waiting for jobs to finish ..', file=sys.stderr)
    for i in range(job_count):
        job_semaphore.acquire()

    output.sort()
    for l in output:
        print(l)

    print('''
#if defined(__GNUG__)
#pragma GCC diagnostic pop
#endif
''')



#!/usr/bin/env python

import argparse
import os
import sys
import subprocess

DEFAULT_MRPT_VERSION = '1:1.3.0-1'

# args
parser = argparse.ArgumentParser()
parser.add_argument('-b', '--build_dir', help='Path to the build directory.')
parser.add_argument('-m', '--mrpt_version',help='The MRPT lib version those bindings are generated from. DEFAULT: {}'.format(DEFAULT_MRPT_VERSION))
parser.add_argument('-a', '--architecture', required=True, help='The architecture the bindings are built for (i386, amd, armhf, etc.).')
args = parser.parse_args()


# check requirements
def check_required_program(program):
    try:
        subprocess.call(['which', program])
    except subprocess.CalledProcessError:
        print 'Required program "{}" not found.'.format(program)
        sys.exit(1)


print 'Looking for required programs:'
check_required_program('sudo')
check_required_program('dpkg')
check_required_program('chrpath')
check_required_program('lintian')

# get supplied build path, otherwise set default
if args.build_dir:
    build_dir = args.build_dir
else:
    build_dir = os.path.join(
      os.path.dirname(os.path.realpath(__file__)), '..', 'build')

# save current dir
curr_dir = str(os.path.abspath(os.path.curdir))

# find pymrpt.so in build dir
built_lib = os.path.join(build_dir, 'lib', 'pymrpt.so')
if os.path.exists(os.path.join(built_lib)):
    print 'Found local build: "{}".'.format(built_lib)
else:
    print 'Could not find "pymrpt.so" in {}. Supply path to build directory as argument!'.format(build_dir)
    print 'Example: {} <path/to/build>'.format(sys.argv[0])
    print '\nExit.'
    sys.exit(1)

# TODO find version automatically
built_lib_version = args.mrpt_version if args.mrpt_version else DEFAULT_MRPT_VERSION
built_lib_size = os.path.getsize(built_lib)
built_lib_arch = args.architecture

# user home
home_dir = os.path.expanduser('~')

# packaging dir
pkg_dir = os.path.join(home_dir, 'mrpt-python-bindings')

# check if packaging dir exists
if os.path.exists(pkg_dir):
    cont = raw_input('Packaging dir already exists. Continue anyway? (y/N): ')
    if cont != 'y':
        print '\nExit.'
        sys.exit(0)
    # remove existing dir (utilizing sudo)
    try:
        command = 'rm -rf {}'.format(pkg_dir)
        subprocess.call(["/usr/bin/sudo", "sh", "-c", command])
    except subprocess.CalledProcessError:
        raise
    print 'Removed existing packaging dir.'

print 'Preparing:'
print '========='

# create packaging directory
os.mkdir(pkg_dir)
print 'Created packaging dir: "{}".'.format(pkg_dir)

# create DEBIAN directory
deb_dir = os.path.join(pkg_dir, 'DEBIAN')
os.mkdir(deb_dir)
print 'Created DEBIAN dir: "{}".'.format(deb_dir)

# create install directory (utilizing sudo)
inst_dir = os.path.join(pkg_dir, 'usr', 'lib', 'python2.7', 'dist-packages')
try:
    command = 'mkdir -p {}'.format(inst_dir)
    subprocess.call(["/usr/bin/sudo", "sh", "-c", command])
except subprocess.CalledProcessError:
    raise
print 'Created install dir: "{}".'.format(inst_dir)

# copy local built pymrpt.so to install dir (utilizing sudo)
command = 'cp {} {}'.format(built_lib, inst_dir)
subprocess.call(["/usr/bin/sudo", "sh", "-c", command])
print 'Copied library file: "{}" to "{}".'.format(built_lib, inst_dir)

# change file permissions
inst_so = os.path.join(inst_dir, 'pymrpt.so')
command = 'chmod 0644 {}'.format(inst_so)
subprocess.call(["/usr/bin/sudo", "sh", "-c", command])
print 'Changed "{}" permissions to: 0644.'.format(inst_so)

# remove rpath from shared lib
command = 'chrpath -d {}'.format(inst_so)
subprocess.call(["/usr/bin/sudo", "sh", "-c", command])
print 'Removed RPATH from "{}".'.format(inst_so)


# create control file
control_content = [
    'Package: mrpt-python-bindings',
    'Version: {}'.format(built_lib_version),
    'Architecture: {}'.format(built_lib_arch),
    'Maintainer: Peter Rudolph <semael23@gmail.com>',
    'Installed-Size: {}'.format(built_lib_size),
    'Depends: libmrpt-base1.3, libmrpt-slam1.3, libmrpt-obs1.3, libmrpt-maps1.3, libmrpt-nav1.3, libmrpt-opengl1.3, libmrpt-gui1.3, libboost-python1.54, libc6',
    'Section: devel',
    'Priority: optional',
    'Homepage: https://www.mrpt.org',
    'Description: MRPT python bindings.',
    ' This package contains the python bindings for',
    ' the Mobile Robot Programming Toolkit (MRPT).',
    ''  # final new line
]
control_filename = os.path.join(pkg_dir, 'DEBIAN', 'control')
control_file = open(control_filename, 'w+')
control_file.write('\n'.join(control_content))
control_file.close()
print 'Created control file: "{}".'.format(control_filename)


# create copyright file
copyright_content = [
    'Format: http://www.debian.org/doc/packaging-manuals/copyright-format/1.0/',
    'Upstream-Name: mrpt-python-bindings',
    'Upstream-Contact: Peter Rudolph <semael23@gmail.com>',
    'Source: https://github.com/MRPT/mrpt',
    '',
    'Files: *',
    'Copyright: 2014-2015 Peter Rudolph',
    'License: Expat',
    '',
    'License: Expat',
    ' Permission is hereby granted, free of charge, to any person obtaining a copy',
    ' of this software and associated documentation files (the "Software"), to deal',
    ' in the Software without restriction, including without limitation the rights',
    ' to use, copy, modify, merge, publish, distribute, sublicense, and/or sell',
    ' copies of the Software, and to permit persons to whom the Software is',
    ' furnished to do so, subject to the following conditions:',
    ' ',
    ' The above copyright notice and this permission notice shall be included in',
    ' all copies or substantial portions of the Software.',
    ' ',
    ' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR',
    ' IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,',
    ' FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE',
    ' AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER',
    ' LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,',
    ' OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN',
    ' THE SOFTWARE.',
    '',
]
copyright_filename = os.path.join(pkg_dir, 'DEBIAN', 'copyright')
copyright_file = open(copyright_filename, 'w+')
copyright_file.write('\n'.join(copyright_content))
copyright_file.close()
print 'Created copyright file: "{}".'.format(copyright_filename)

# TODO use auto-generated changelog
changelog_content = [
    'python-mrpt-bindings ({}) unstable; urgency=low'.format(built_lib_version),
    '',
    '  * Initial Release.',
    '',
    ' -- Peter Rudolph <semael23@gmail.com>  Tue, 20 Jan 2015 12:30:54 +0100',
    '',
]
changelog_filename = os.path.join(pkg_dir, 'DEBIAN', 'changelog')
changelog_file = open(changelog_filename, 'w+')
changelog_file.write('\n'.join(changelog_content))
changelog_file.close()
print 'Created changelog file: "{}".'.format(changelog_filename)


os.chdir(home_dir)

# build the package
print 'Build package:'
subprocess.call(['dpkg', '-b', 'mrpt-python-bindings'])

# check package with lintian
print 'Check package:'
subprocess.call(
  ['lintian', os.path.join(home_dir, 'mrpt-python-bindings.deb')])

# go to initial dir
os.chdir(curr_dir)

print 'Done.'



#!/usr/bin/env python3
import sys
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import timeit

import mpld3
from mpld3 import plugins, utils

# Install tabulate using "pip-install tabulate"
from tabulate import tabulate

import os


def clear(): os.system('clear')
# Function to clear screen
import pymrpt


# Matplotlib figure parameters
mpl.rcParams['figure.figsize'] = (12.0, 8.0)
plt.rcParams.update({'axes.titlesize': 30, 'axes.labelsize': 20, 'xtick.labelsize': 15, 'ytick.labelsize': 15,
                     'figure.titlesize': 40})

# Define number of points and camera parameters
n = 10
sigma = 0.0005
n_range = list(range(5, 25))
sigma_range = np.arange(0.0001, 0.001, 0.0001)
f = 1.0
cx = 0.0
cy = 0.0
cam_intrinsic = np.array([[f, 0.0, cx], [0.0, f, cy], [0.0, 0.0, 1.0]])

# Define constants for serial outputting results
checkmark = '\u2713'
l_progress_bar = 50

# Instantiate pnp module
pnp = pymrpt.pnp(n)

# Define settings for comparison module
algos = [pnp.dls, pnp.epnp, pnp.p3p,
         pnp.rpnp, pnp.ppnp, pnp.posit, pnp.lhm]
algo_names = ['dls', 'epnp', 'p3p', 'rpnp', 'ppnp', 'posit', 'lhm']
algo_ls = [':', '-', '--', '-', '--', '-', '-']
n_algos = len(algos)
n_iter = 100


class HighlightLines(plugins.PluginBase):
    # css format for interactive d3 plots
    """A plugin for an interactive legend.
    Inspired by http://bl.ocks.org/simzou/6439398
    """

    JAVASCRIPT = """
    mpld3.register_plugin("interactive_legend", InteractiveLegend);
    InteractiveLegend.prototype = Object.create(mpld3.Plugin.prototype);
    InteractiveLegend.prototype.constructor = InteractiveLegend;
    InteractiveLegend.prototype.requiredProps = ["line_ids", "labels"];
    InteractiveLegend.prototype.defaultProps = {}
    function InteractiveLegend(fig, props){
        mpld3.Plugin.call(this, fig, props);
    };
    InteractiveLegend.prototype.draw = function(){
        var labels = new Array();
        for(var i=0; i<this.props.labels.length; i++){
            var obj = {}
            obj.label = this.props.labels[i]
            obj.line = mpld3.get_element(this.props.line_ids[i], this.fig)
            obj.visible = false;
            labels.push(obj);
        }
        var ax = this.fig.axes[0]
        var legend = this.fig.canvas.append("svg:g")
                               .attr("class", "legend");
        // add the rectangles
        legend.selectAll("rect")
                .data(labels)
             .enter().append("rect")
                .attr("height",10)
                .attr("width", 25)
                .attr("x",ax.width+10+ax.position[0])
                .attr("y",function(d,i) {
                                return ax.position[1]+ i * 25 - 10;})
                .attr("stroke", function(d) {
                                return d.line.props.edgecolor})
                .attr("class", "legend-box")
                .style("fill", "white")
                .on("click", click)
        // add the text
        legend.selectAll("text")
                .data(labels)
            .enter().append("text")
              .attr("x", function (d) {
                return ax.width+10+ax.position[0] + 25 + 15
              })
              .attr("y", function(d,i) {
                return ax.position[1]+ i * 25
              })
              .text(function(d) { return d.label })
        // specify the action on click
        function click(d,i){
            d.visible = !d.visible;
            d3.select(this)
              .style("fill",function(d, i) {
                  console.log(d)
                  var color = d.line.props.edgecolor
                  return d.visible ? color : "white";
              })
            d3.select(d.line.path[0][0])
                .style("stroke-opacity", d.visible ? 1 : d.line.props.alpha);
        }
    };
    """

    def __init__(self, lines, labels, css):

        self.css_ = css or ""

        self.lines = lines
        self.dict_ = {"type": "interactive_legend",
                      "line_ids": [utils.get_id(line) for line in lines],
                      "labels": labels}


css = """
.legend-box {
  cursor: pointer;
}
"""


def vector2RotMat(vec, theta=0):
    # Function to convert from axis, angle to rotation matrix
    # Rodrigues rotation formula
    n_check = np.linalg.norm(vec)

    kx = vec[0] / n_check
    ky = vec[1] / n_check
    kz = vec[2] / n_check

    K = np.matrix([[0, -kz, ky], [kz, 0, -kx], [-ky, kx, 0]])
    I = np.identity(3)

    R = I + K * np.sin(theta) + K * K * (1 - np.cos(theta))
    R = np.array(R)

    return R


def quatvec2RotMat(q):
    # Function to convert from quaternion to Rotaiton matrix
    qw = np.sqrt(1 - np.linalg.norm(q) * np.linalg.norm(q))
    qx = q[0]
    qy = q[1]
    qz = q[2]

    R = [1 - 2 * qy * qy - 2 * qz * qz, 2 * qx * qy - 2 * qz * qw, 2 * qx * qz + 2 * qy * qw,
         2 * qx * qy + 2 * qz * qw, 1 - 2 * qx * qx - 2 * qz * qz, 2 * qy * qz - 2 * qx * qw,
         2 * qx * qz - 2 * qy * qw, 2 * qy * qz + 2 * qx * qw, 1 - 2 * qx * qx - 2 * qy * qy]

    R = np.reshape(R, [3, 3])

    return R


def RotMat2quat(R):
    # Function to convert from rotation matrix to Quaternion
    qw = np.sqrt(1 + R[0, 0] + R[1, 1] + R[2, 2]) / 2

    if qw > 0.01:
        qx = (R[2, 1] - R[1, 2]) / 4 / qw
        qy = (R[0, 2] - R[2, 0]) / 4 / qw
        qz = (R[1, 0] - R[0, 1]) / 4 / qw
    else:
        l = np.array([R[0, 0], R[1, 1], R[2, 2]])
        ind_max = np.argmax(l)

        if ind_max == 0:
            qx = np.sqrt((R[0, 0] + 1) / 2)
            qy = (R[1, 0] + R[0, 1]) / 4 / qx
            qz = (R[0, 2] + R[2, 0]) / 4 / qx

        elif ind_max == 1:
            qy = np.sqrt((R[1, 1] + 1) / 2)
            qx = (R[1, 0] + R[0, 1]) / 4 / qy
            qz = (R[2, 1] + R[1, 2]) / 4 / qy

        else:
            qz = np.sqrt((R[2, 2] + 1) / 2)
            qx = (R[0, 2] + R[2, 0]) / 4 / qz
            qy = (R[2, 1] + R[1, 2]) / 4 / qz

        qw = np.sqrt(1 - qx * qx - qy * qy - qz * qz)

    return [qw, qx, qy, qz]


def display_comparison_plot(t, arr, names, line_styles, title, xtitle, ytitle, ylim, figname):

    f, ax = plt.subplots()
    lines = []
    for i in np.arange(0, len(names)):
        l, = ax.plot(t, arr[:, i], label=names[i], lw=3, ls=line_styles[i])
        lines.append(l)

    leg = ax.legend(fancybox=True, shadow=True)
    leg.get_frame().set_alpha(0.8)

    lined = dict()
    for legline, origline in zip(leg.get_lines(), lines):
        legline.set_picker(10)
        lined[legline] = origline

    ax.set_xlabel(xtitle)
    ax.set_ylabel(ytitle)
    ax.set_title(title)
    ax = plt.gca()
    ax.set_ylim(ylim)
    ax.grid()

    def onpick(event):
        legline = event.artist
        origline = lined[legline]
        vis = not origline.get_visible()
        origline.set_visible(vis)

        if vis:
            legline.set_alpha(1)
        else:
            legline.set_alpha(0.2)

        f.canvas.draw()

    f.canvas.mpl_connect('pick_event', onpick)

    plt.show()
    plt.savefig(figname + '.pdf')


def display_comparison_plot_mpld3(t, arr, names, line_styles, title, xtitle, ytitle, ylim, figname):
    # Function used to generate interactive d3 plots in html
    f, ax = plt.subplots()
    lines = []
    for i in np.arange(0, len(names)):
        l, = ax.plot(t, arr[:, i], label=names[i], lw=3, ls=line_styles[i], alpha=0.2)
        lines.append(l)

    ax.set_xlabel(xtitle)
    ax.set_ylabel(ytitle)
    ax.set_title(title)
    ax = plt.gca()
    ax.set_ylim(ylim)
    ax.grid()

    plugins.connect(f, HighlightLines(lines, names, css))

    mpld3.display()

    #mpld3.save_html(f, figname + '.html')

    return mpld3.fig_to_html(f)


def printProgress(iteration, total, prefix='', suffix='', decimals=1, barLength=100):
    """
    # Print iterations progress
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        barLength   - Optional  : character length of bar (Int)
    """
    formatStr = "{0:." + str(decimals) + "f}"
    percents = formatStr.format(100 * (iteration / float(total)))
    filledLength = int(round(barLength * iteration / float(total)))
    bar = "\u2588" * filledLength + '-' * (barLength - filledLength)
    sys.stdout.write('\r%s |%s| %s%s %s' % (prefix, bar, percents, '%', suffix)),
    sys.stdout.flush()
    if iteration == total:
        sys.stdout.write('\n')
        sys.stdout.flush()


def printTestStatus(test_index):
    # Print overall progress of all the tests
    head_tests = ['#No.', 'Status', 'Description']

    table_tests = [[str(1), '', 'Test for 100 iterations with each algorithm'], [str(2), '',  'Varitaion of error with image pixel noise standard deviation (' +
                                                                                 '\u03C3' + ')'], [str(3), '', ' Variation in error with number of 2d/3d correspondences (n)'], [str(4), '', 'Average computation time of each algorithm']]

    for i in range(0, test_index):
        table_tests[i][1] = checkmark

    print(tabulate(table_tests, head_tests, tablefmt='fancy_grid'))

    for i in range(0, test_index):
        printProgress(l_progress_bar, l_progress_bar, prefix='Test' + str(i + 1) +
                      ' Progress:', suffix='Complete', barLength=50)


def printTest1Results(vals):
    # Function to print results of test 1
    test1_headers = ['Algo', 'Translation Mean Error', 'Translation Median Error',
                     'Rotation Mean Error', 'Rotation Median Error']

    test1_table = np.empty([7, 5], dtype=object)
    test1_table[:, 1:] = vals
    test1_table[:, 0] = algo_names

    print(tabulate(test1_table, test1_headers, tablefmt='fancy_grid'))


def printTest4Results(vals):
    # Function to print results of test 4
    test4_headers = ['Algo', 'Translation Mean Error', 'Translation Median Error',
                     'Rotation Mean Error', 'Rotation Median Error']

    vals = np.random.rand(7, 1)
    test4_table = np.empty([7, 2], dtype=object)
    test4_table[:, 1:] = vals
    test4_table[:, 0] = algo_names

    print(tabulate(test4_table, test4_headers, tablefmt='fancy_grid'))


def calc_err(pose1, pose2):
    # FUnction to compute reprojection errors
    if np.any(np.isnan(pose1)) or np.linalg.norm(pose1) > 1000000:
        err = [0, 0]
        return err

    # Percent error in translation
    t_est = np.array(pose1[0:3])
    t = np.array(pose2[0:3])
    err_t = (np.linalg.norm(t_est - t) / np.linalg.norm(t)) * 100
    # Rotation error
    q_est = pose1[3:6, 0]
    q = pose2[3:6, 0]

    if np.linalg.norm(q) != 0 and np.linalg.norm(q_est):
        val = np.dot(q_est, q) / np.linalg.norm(q_est) / np.linalg.norm(q)
    else:
        val = 1

    if val > 1:
        val = 1
    elif val < -1:
        val = -1
    elif val == np.nan:
        val = 1

    err_q = np.max(np.abs(np.arccos(val))) * 180 / np.pi
    err_q = err_q if err_q < 180 - err_q else 180 - err_q

    err = [err_t, err_q]

    if np.isnan(err_t) or np.isnan(err_q):
        print('pose_est=\n', pose2)
        print('pose=\n', pose1)

    return err


def err_plot():
    # Test 1
    # Define object points and image points

    obj_pts = np.random.randint(-40, 40, [n, 3])
    obj_pts = obj_pts.astype(float)
    obj_pts[0, :] = [0, 0, 0]
    img_pts = np.empty([n, 3])
    img_pts[:, 2] = 1

    pose_est = np.empty([6, 1])
    pose_act = np.empty([6, 1])

    err_net_t = []
    err_net_q = []

    for it in np.arange(0, n_iter):

        # Generate random camera extrinsic matrix
        v = 2 * np.random.random([3]) - np.array([1, 1, 1])
        v = v / np.linalg.norm(v)

        #R=np.array([[0.841986, -0.352662, -0.408276],[0.308904,  0.935579, -0.171085],[0.442309, 0.0179335,  0.896683]])
        R = vector2RotMat(v, np.pi * 2 / 3)
        q = RotMat2quat(R)
        t = np.array([0.0, 0.0, 200.0])

        # Compute image points based on actual extrinsic matrix and add noise to measurements
        for i in range(0, n):
            pt = np.dot(R, obj_pts[i, :]) + t
            img_pts[i, 0:2] = np.array([pt[0] / pt[2], pt[1] / pt[2]])

        img_pts[:, 0:2] = img_pts[:, 0:2] + sigma * np.random.randn(n, 2)

        pose_act[0:3, 0] = t
        pose_act[3:6, 0] = q[1:4]

        # Use the c-library to compute the pose
        err_t = []
        err_q = []
        for i in np.arange(0, n_algos):
            algos[i](obj_pts, img_pts, n, cam_intrinsic, pose_est)
            e = calc_err(pose_est, pose_act)
            err_t.append(e[0])
            err_q.append(e[1])

            if np.isnan(e[0]) or np.isnan(e[1]):
                print('algo=\n', algo_names[i])
                print('err=\n', e)

        err_net_t.append(err_t)
        err_net_q.append(err_q)
        printProgress(it, n_iter, prefix='Test' + str(1) +
                      ' Progress:', suffix='Complete', barLength=50)
    mean_err_t = np.mean(err_net_t, axis=0)
    mean_err_q = np.mean(err_net_q, axis=0)
    median_err_t = np.median(err_net_t, axis=0)
    median_err_q = np.median(err_net_q, axis=0)

    vals = np.empty([7, 4])
    vals[:, 0] = mean_err_t
    vals[:, 1] = median_err_t
    vals[:, 2] = mean_err_q
    vals[:, 3] = median_err_q
    """
    for i in np.arange(0, n_algos):
        print 'mean_err_t_' + algo_names[i] + '=', mean_err_t[i], 'median_err_t_' + algo_names[i] + '=', median_err_t[i]
        print 'mean_err_q_' + algo_names[i] + '=', mean_err_q[i], 'median_err_q_' + algo_names[i] + '=', median_err_q[i]
    """
    it = np.arange(0, n_iter)

    err_net_t = np.array(err_net_t)
    err_net_q = np.array(err_net_q)

    s = '<h2> Translation error and Rotation error for 100 iterations (R - Randomly varying, t - fixed) </h2>'

    s1 = display_comparison_plot_mpld3(it, err_net_t, names=algo_names, line_styles=algo_ls,
                                       title='% Translation Error Plot', xtitle='Iteration', ytitle='e_t', ylim=[0, 2], figname='err_t')
    s2 = display_comparison_plot_mpld3(it, err_net_q, names=algo_names, line_styles=algo_ls,
                                       title='Rotation Error Plot (deg)', xtitle='Iteration', ytitle='e_q', ylim=[0, 1], figname='err_q')

    s = s + '\n <table > \n <tr > \n <td > \n' + s1 + \
        '</td > \n <td> \n' + s2 + '</td> \n </tr> \n </table> \n'

    return s, vals

    """
    comp_arr = np.zeros([n_iter,2])
    for i in np.arange(0,n_iter):
        comp_arr[i,:] = 2*mean_err_p3p
    nconv_epnp = float(np.sum(err_epnp>comp_arr))/n_iter*100
    nconv_dls  = float(np.sum(err_dls>comp_arr))/n_iter*100
    nconv_ppnp  = float(np.sum(err_ppnp>comp_arr))/n_iter*100
    nconv_posit  = float(np.sum(err_posit>comp_arr))/n_iter*100
    nconv_lhm  = float(np.sum(err_lhm>comp_arr))/n_iter*100
    nconv_p3p  = float(np.sum(err_p3p>comp_arr))/n_iter*100
    plt.figure()
    xvals = ['epnp', 'dls', 'ppnp', 'posit','lhm', 'p3p']
    xvals_int = np.arange(0,n_algos)
    yvals = [nconv_epnp, nconv_dls, nconv_ppnp, nconv_posit, nconv_lhm, nconv_p3p]
    plt.bar(xvals_int, yvals, align='center')
    plt.xticks(xvals_int, xvals)
    plt.title('%Divergence')
    plt.show()
    plt.savefig('divergence.pdf')
    """


def err_statistics_fcn_n():
    # Test 2
    mean_err_t_net = []
    mean_err_q_net = []
    median_err_t_net = []
    median_err_q_net = []

    for n in n_range:
        # Define object points and image points
        obj_pts = np.random.randint(-40, 40, [n, 3])
        obj_pts = obj_pts.astype(float)
        obj_pts[0, :] = [0, 0, 0]
        img_pts = np.empty([n, 3])
        img_pts[:, 2] = 1

        pose_est = np.empty([6, 1])
        pose_act = np.empty([6, 1])

        err_net_t = []
        err_net_q = []

        for it in np.arange(0, n_iter):

            # Define camera extrinsic matrix
            v = 2 * np.random.random([3]) - np.array([1, 1, 1])
            v = v / np.linalg.norm(v)

            #R=np.array([[0.841986, -0.352662, -0.408276],[0.308904,  0.935579, -0.171085],[0.442309, 0.0179335,  0.896683]])
            R = vector2RotMat(v, np.pi * 2 / 3)
            q = RotMat2quat(R)
            t = np.array([0.0, 0.0, 200.0])

            # Compute image points based on actual extrinsic matrix and add noise to measurements
            for i in range(0, n):
                pt = np.dot(R, obj_pts[i, :]) + t
                img_pts[i, 0:2] = np.array([pt[0] / pt[2], pt[1] / pt[2]])

            img_pts[:, 0:2] = img_pts[:, 0:2] + sigma * np.random.randn(n, 2)

            pose_act[0:3, 0] = t
            pose_act[3:6, 0] = q[1:4]

            # Use the c-library to compute the pose

            err_t = []
            err_q = []
            for i in range(0, n_algos):
                algos[i](obj_pts, img_pts, n, cam_intrinsic, pose_est)
                e = calc_err(pose_est, pose_act)
                err_t.append(e[0])
                err_q.append(e[1])

            err_net_t.append(err_t)
            err_net_q.append(err_q)

        mean_err_t_net.append(np.mean(err_net_t, axis=0))
        mean_err_q_net.append(np.mean(err_net_q, axis=0))

        median_err_t_net.append(np.median(err_net_t, axis=0))
        median_err_q_net.append(np.median(err_net_q, axis=0))

        printProgress(n - 5, len(n_range), prefix='Test' + str(3) +
                      ' Progress:', suffix='Complete', barLength=50)

    it = np.arange(5, 25)

    mean_err_t_net = np.array(mean_err_t_net)
    mean_err_q_net = np.array(mean_err_q_net)
    median_err_t_net = np.array(median_err_t_net)
    median_err_q_net = np.array(median_err_q_net)

    s = '<h2> Mean and Median error in Translation and Rotation with varying 2d/3d correspondences (n) </h2>'

    s1 = display_comparison_plot_mpld3(it, mean_err_t_net, names=algo_names, line_styles=algo_ls,
                                       title='Mean Translation %Error Plot', xtitle='n', ytitle=r'% Translation error e_t', ylim=[0, 10], figname='mean_err_t')
    s2 = display_comparison_plot_mpld3(it, mean_err_q_net, names=algo_names, line_styles=algo_ls,
                                       title='Mean Rotation Error Plot (deg)', xtitle='n', ytitle=r'Rotation error e_q (deg)', ylim=[0, 0.5], figname='mean_err_q')

    s3 = display_comparison_plot_mpld3(it, median_err_t_net, names=algo_names, line_styles=algo_ls,
                                       title='Median Translation %Error Plot', xtitle='n', ytitle=r'% Translation error e_t', ylim=[0, 1], figname='median_err_t')
    s4 = display_comparison_plot_mpld3(it, median_err_q_net, names=algo_names, line_styles=algo_ls,
                                       title='Median Rotation Error Plot (deg)', xtitle='n', ytitle=r'Rotation error e_q(deg)', ylim=[0, 0.5], figname='median_err_q')

    s = s + '\n<table>\n <tr>\n <td>\n' + s1 + '</td>\n <td>\n' + s2 + '</td>\n </tr>\n' + \
        '\n<tr>\n <td>\n' + s3 + '\n</td>\n <td>\n' + s4 + '\n</td>\n </tr>\n </table>\n'
    return s


def err_statistics_fcn_sigma():
    # Test 3
    mean_err_t_net = []
    mean_err_q_net = []
    median_err_t_net = []
    median_err_q_net = []

    n = 10

    for sigma in sigma_range:
        # Define object points and image points
        obj_pts = np.random.randint(-40, 40, [n, 3])
        obj_pts = obj_pts.astype(float)
        obj_pts[0, :] = [0, 0, 0]
        img_pts = np.empty([n, 3])
        img_pts[:, 2] = 1

        pose_est = np.empty([6, 1])
        pose_act = np.empty([6, 1])

        err_net_t = []
        err_net_q = []

        for it in np.arange(0, n_iter):

            # Define camera extrinsic matrix
            v = 2 * np.random.random([3]) - np.array([1, 1, 1])
            v = v / np.linalg.norm(v)

            #R=np.array([[0.841986, -0.352662, -0.408276],[0.308904,  0.935579, -0.171085],[0.442309, 0.0179335,  0.896683]])
            R = vector2RotMat(v, np.pi * 2 / 3)
            q = RotMat2quat(R)
            t = np.array([0.0, 0.0, 200.0])

            # Compute image points based on actual extrinsic matrix and add noise to measurements
            for i in range(0, n):
                pt = np.dot(R, obj_pts[i, :]) + t
                img_pts[i, 0:2] = np.array([pt[0] / pt[2], pt[1] / pt[2]])

            img_pts[:, 0:2] = img_pts[:, 0:2] + sigma * np.random.randn(n, 2)

            pose_act[0:3, 0] = t
            pose_act[3:6, 0] = q[1:4]

            # Use the c-library to compute the pose
            err_t = []
            err_q = []
            for i in range(0, n_algos):
                algos[i](obj_pts, img_pts, n, cam_intrinsic, pose_est)
                e = calc_err(pose_est, pose_act)
                err_t.append(e[0])
                err_q.append(e[1])

            err_net_t.append(err_t)
            err_net_q.append(err_q)

        mean_err_t_net.append(np.mean(err_net_t, axis=0))
        mean_err_q_net.append(np.mean(err_net_q, axis=0))

        median_err_t_net.append(np.median(err_net_t, axis=0))
        median_err_q_net.append(np.median(err_net_q, axis=0))

        printProgress(sigma * 1000, len(sigma_range), prefix='Test' + str(2) +
                      ' Progress:', suffix='Complete', barLength=50)

    it = np.arange(0.001, 0.010, 0.001)

    mean_err_t_net = np.array(mean_err_t_net)
    mean_err_q_net = np.array(mean_err_t_net)

    median_err_t_net = np.array(median_err_t_net)
    median_err_q_net = np.array(median_err_t_net)

    s = '\n<h2>\n Mean and Median error in Translation and Rotation with varying noise variance (sigma)\n </h2>\n'

    s1 = display_comparison_plot_mpld3(it, mean_err_t_net, names=algo_names, line_styles=algo_ls, title='Mean Translation %Error Plot',
                                       xtitle=r'\sigma', ytitle=r'% Translation error e_t', ylim=[0, 10], figname='mean_sigma_err_t')
    s2 = display_comparison_plot_mpld3(it, mean_err_q_net, names=algo_names, line_styles=algo_ls, title='Mean Rotation Error Plot (deg)',
                                       xtitle=r'\sigma', ytitle=r'Rotation error e_q (deg)', ylim=[0, 0.5], figname='mean_sigma_err_q')

    s3 = display_comparison_plot_mpld3(it, median_err_t_net, names=algo_names, line_styles=algo_ls, title='Median Translation %Error Plot',
                                       xtitle=r'\sigma', ytitle=r'% Translation error e_t', ylim=[0, 1], figname='median_sigma_err_t')
    s4 = display_comparison_plot_mpld3(it, median_err_q_net, names=algo_names, line_styles=algo_ls, title='Median Rotation Error Plot (deg)',
                                       xtitle=r'\sigma', ytitle=r'Rotation error e_q (deg)', ylim=[0, 0.5], figname='median_sigma_err_q')

    s = s + '\n<table>\n <tr>\n <td>\n' + s1 + '\n</td>\n <td>\n' + s2 + '\n</td>\n </tr>\n' + \
        '\n<tr>\n <td>\n' + s3 + '\n</td> \n<td>\n' + s4 + '\n</td>\n </tr>\n </table>\n'

    return s


def time_comp():
    # Test 4
    obj_pts_store = []
    img_pts_store = []
    n_max = 50
    n_step = 1
    n_start = 10
    n_iter = 10

    tcomp_storage = []

    for n in np.arange(n_start, n_max, n_step):
        # Generate object points and image points
        for i in np.arange(0, n_iter):
            obj_pts = np.random.randint(-40, 40, [n, 3])
            obj_pts = obj_pts.astype(float)
            obj_pts[0, :] = [0, 0, 0]
            img_pts = np.empty([n, 3])
            img_pts[:, 2] = 1

            # Define camera extrinsic matrix
            v = 2 * np.random.random([3]) - np.array([1, 1, 1])
            v = v / np.linalg.norm(v)

            #R=np.array([[0.841986, -0.352662, -0.408276],[0.308904,  0.935579, -0.171085],[0.442309, 0.0179335,  0.896683]])
            R = vector2RotMat(v, np.pi * 2 / 3)
            t = np.array([0.0, 0.0, 200.0])

            # Compute image points based on actual extrinsic matrix
            for i in range(0, n):
                pt = np.dot(R, obj_pts[i, :]) + t
                img_pts[i, 0:2] = np.array([pt[0] / pt[2], pt[1] / pt[2]])

            # Add noise to measurements
            img_pts[:, 0:2] = img_pts[:, 0:2] + sigma * np.random.randn(n, 2)

            obj_pts_store.append(obj_pts)
            img_pts_store.append(img_pts)

        tcomp = []
        # Compute time for n_iter iterations
        for it in np.arange(1, n_algos):
            pose_est = np.empty([6, 1])
            start = timeit.default_timer()
            for i in np.arange(0, n_iter):
                obj_pts = obj_pts_store[i]
                img_pts = img_pts_store[i]
                algos[it](obj_pts, img_pts, n, cam_intrinsic, pose_est)

            end = timeit.default_timer()

            tcomp.append((end - start) / float(n_iter) * 1000.0)

        tcomp_storage.append(tcomp)

        printProgress(n - n_start, len(list(range(n_start, n_max, n_step))), prefix='Test' + str(4) +
                      ' Progress:', suffix='Complete', barLength=50)
    it = np.arange(n_start, n_max, n_step)
    tcomp_storage = np.array(tcomp_storage)

    s = '\n<h2>\n Average compuational time for algorithm (ms) \n</h2>\n'
    s = s + display_comparison_plot_mpld3(it, tcomp_storage, names=algo_names[1:], line_styles=algo_ls[
        1:], title='Average Time for algorithm (ms)', xtitle=r'n', ytitle=r't(ms)', ylim=[0, 0.6], figname='mean_time')

    return s

# Introduction and links to various files
ss = """<!DOCTYPE html>
<html>
<body bgcolor="#E6E6FA">
<br>
<CENTER>
<embed src = "https://www.dropbox.com/s/a266gqe3o0typpg/pnp_intro1.png?raw=1 #toolbar=0&navpanes=0&scrollbar=0" width = "1000" height = "1250" ALIGN=CENTER>
<br>
<embed src = "https://www.dropbox.com/s/5v8n4edc7g8q8tu/pnp_intro2.png?raw=1 #toolbar=0&navpanes=0&scrollbar=0" width = "1000" height = "1250" ALIGN=CENTER>
<br>
<embed src = "https://www.dropbox.com/s/6bdadtn99tthth0/pnp_intro3.png?raw=1 #toolbar=0&navpanes=0&scrollbar=0" width = "1000" height = "650" ALIGN=CENTER>
<br>
<h1> Sample Pose Estimation using Camera Calib application of MRPT </h1>
<br>
<iframe src = "https://www.youtube.com/embed/aGd7ZyrcwaE" width = "960" height = "540" frameborder = "0" allowfullscreen ALIGN=CENTER> </iframe >

<br>
<br>
<h1> Performance Comparison using python interface in MRPT(pnp_perf_comp.py) </h1>
<div>
<h3> Plots are interactive <br>
    * Click on legend box to solidy the particular algorithm <br>
    * Pan and Zoom at lower left corner  </h3>
</div>"""

# sys.stderr.write('\x1b[2J\x1b[H')
clear()
printTestStatus(0)
printProgress(0, len(list(range(5, 25))), prefix='Test' + str(1) +
              ' Progress:', suffix='Complete', barLength=50)
s1, vals = err_plot()

# sys.stderr.write('\x1b[2J\x1b[H')
clear()
printTestStatus(1)
printProgress(0, len(list(range(5, 25))), prefix='Test' + str(2) +
              ' Progress:', suffix='Complete', barLength=50)
s2 = err_statistics_fcn_sigma()

# sys.stderr.write('\x1b[2J\x1b[H')
clear()
printTestStatus(2)
printProgress(0, len(list(range(5, 25))), prefix='Test' + str(3) +
              ' Progress:', suffix='Complete', barLength=50)
s3 = err_statistics_fcn_n()

# sys.stderr.write('\x1b[2J\x1b[H')
clear()
printTestStatus(3)
printProgress(0, len(list(range(5, 25))), prefix='Test' + str(4) +
              ' Progress:', suffix='Complete', barLength=50)
s4 = time_comp()

# sys.stderr.write('\x1b[2J\x1b[H')
clear()
printTestStatus(4)
print('\n\nResults of Test 1 \n\n')
printTest1Results(vals)

s5 = """<h1> MRPT Merge Pull Request </h1>
  <h3>
  <a href="https://github.com/MRPT/mrpt/pull/310">Link to PnP Algorithm Pull Request </a>
  </h3>
</CENTER>
</body>
</html>
"""

ss = ss + s1 + s2 + s3 + s4 + s5

f = open("pnp_perf_comp.html", "w")
f.write(ss)
f.close()



# python/example4.py -- Python version of an example application that
# shows how to use the GLCanvas widget. For a C++ implementation, see
# '../src/example4.cpp'.
#
# NanoGUI was developed by Wenzel Jakob <wenzel@inf.ethz.ch>.
# The widget drawing code is based on the NanoVG demo application
# by Mikko Mononen.
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE.txt file.

import nanogui
import random
import math
import time
import gc

from nanogui import Color, Screen, Window, GroupLayout, BoxLayout, \
                    ToolButton, Label, Button, Widget, \
                    PopupButton, CheckBox, MessageDialog, VScrollPanel, \
                    ImagePanel, ImageView, ComboBox, ProgressBar, Slider, \
                    TextBox, ColorWheel, Graph, GridLayout, \
                    Alignment, Orientation, TabWidget, IntBox, GLShader, GLCanvas

from nanogui import gl, glfw, entypo


class MyGLCanvas(GLCanvas):
    def __init__(self, parent):
        super(MyGLCanvas, self).__init__(parent)

        try:
            import numpy as np

            self.shader = GLShader()
            self.shader.init(
                # An identifying name
                "a_simple_shader",

                # Vertex shader
                """#version 330
                uniform mat4 modelViewProj;
                in vec3 position;
                in vec3 color;
                out vec4 frag_color;
                void main() {
                    frag_color = 3.0 * modelViewProj * vec4(color, 1.0);
                    gl_Position = modelViewProj * vec4(position, 1.0);
                }""",

                # Fragment shader
                """#version 330
                out vec4 color;
                in vec4 frag_color;
                void main() {
                    color = frag_color;
                }"""
            )

            # Draw a cube
            indices = np.array(
                [[0, 3, 3, 6, 7, 5, 4, 1, 4, 3, 5, 2],
                 [1, 2, 2, 7, 6, 4, 5, 0, 0, 7, 6, 1],
                 [3, 1, 6, 3, 5, 7, 1, 4, 3, 4, 2, 5]],
                dtype=np.int32)

            positions = np.array(
                [[-1, -1,  1,  1, -1, -1,  1,  1],
                 [ 1,  1,  1,  1, -1, -1, -1, -1],
                 [ 1, -1, -1,  1,  1, -1, -1,  1]],
                dtype=np.float32)

            colors = np.array(
                [[1, 0, 1, 0, 1, 0, 1, 0.5, 1, 1, 0.5, 0.5],
                 [0, 1, 1, 0, 0, 1, 1, 0.5, 0, 0.5, 1, 1],
                 [0, 0, 0, 1, 1, 1, 1, 0.5, 0.5, 0, 0, 0.5]],
                dtype=np.float32)

            self.shader.bind()
            self.shader.uploadIndices(indices)

            self.shader.uploadAttrib("position", positions)
            self.shader.uploadAttrib("color", colors)
        except ImportError:
            self.shader = None
            pass

    def drawGL(self):
        if self.shader is not None:
            import numpy as np
            self.shader.bind()

            current_time = time.time()
            angle_x = self.rotation[0] * current_time
            angle_y = self.rotation[1] * current_time
            angle_z = self.rotation[2] * current_time

            mvp_rotX = np.matrix(
                [[1, 0, 0, 0],
                 [0, np.cos(angle_x), -np.sin(angle_x), 0],
                 [0, np.sin(angle_x), np.cos(angle_x), 0],
                 [0, 0, 0, 1]],
                dtype=np.float32)

            mvp_rotY = np.matrix(
                [[np.cos(angle_y), 0, np.sin(angle_y), 0],
                 [0, 1, 0, 0],
                 [-np.sin(angle_y), 0, np.cos(angle_y), 0],
                 [0, 0, 0, 1]],
                dtype=np.float32)

            mvp_rotZ = np.matrix(
                [[np.cos(angle_z), -np.sin(angle_z), 0, 0],
                 [np.sin(angle_z), np.cos(angle_z), 0, 0],
                 [0, 0, 1, 0],
                 [0, 0, 0, 1]],
                dtype=np.float32)

            mvp = mvp_rotX * mvp_rotY * mvp_rotZ

            mvp[0:3, 0:3] *= 0.25

            self.shader.setUniform("modelViewProj", mvp)

            gl.Enable(gl.DEPTH_TEST)
            self.shader.drawIndexed(gl.TRIANGLES, 0, 12)
            gl.Disable(gl.DEPTH_TEST)


class TestApp(Screen):
    def __init__(self):
        super(TestApp, self).__init__((800, 600), "NanoGUI Test", False)

        window = Window(self, "GLCanvas Demo")
        window.setPosition((15, 15))
        window.setLayout(GroupLayout())

        self.canvas = MyGLCanvas(window)
        self.canvas.setBackgroundColor(Color(0.5, 0.5, 0.5, 1.0))
        self.canvas.setSize((400, 400))

        self.canvas.rotation = [0.25, 0.5, 0.33]

        tools = Widget(window)
        tools.setLayout(BoxLayout(Orientation.Horizontal,
                                  Alignment.Middle, 0, 5))

        b0 = Button(tools, "Random Color")
        def cb0():
            self.canvas.setBackgroundColor(Color(random.random(), random.random(), random.random(), 1.0))
        b0.setCallback(cb0)

        b1 = Button(tools, "Random Rotation")
        def cb1():
            self.canvas.rotation = [random.random(), random.random(), random.random()]
        b1.setCallback(cb1)

        self.performLayout()

    def keyboardEvent(self, key, scancode, action, modifiers):
        if super(TestApp, self).keyboardEvent(key, scancode,
                                              action, modifiers):
            return True
        if key == glfw.KEY_ESCAPE and action == glfw.PRESS:
            self.setVisible(False)
            return True
        return False

if __name__ == "__main__":
    nanogui.init()
    test = TestApp()
    test.drawAll()
    test.setVisible(True)
    nanogui.mainloop()
    del test
    gc.collect()
    nanogui.shutdown()



# python/example1.py -- Python version of an example application that shows
# how to use the various widget classes. For a C++ implementation, see
# '../src/example1.cpp'.
#
# NanoGUI was developed by Wenzel Jakob <wenzel@inf.ethz.ch>.
# The widget drawing code is based on the NanoVG demo application
# by Mikko Mononen.
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE.txt file.

import nanogui
import math
import time
import gc

from nanogui import Color, ColorPicker, Screen, Window, GroupLayout, BoxLayout, \
                    ToolButton, Label, Button, Widget, \
                    Popup, PopupButton, CheckBox, MessageDialog, VScrollPanel, \
                    ImagePanel, ImageView, ComboBox, ProgressBar, Slider, \
                    TextBox, ColorWheel, Graph, GridLayout, \
                    Alignment, Orientation, TabWidget, IntBox, GLShader

from nanogui import gl, glfw, entypo

# A simple counter, used for dynamic tab creation with TabWidget callback
counter = 1

class TestApp(Screen):
    def __init__(self):
        super(TestApp, self).__init__((1024, 768), "NanoGUI Test")

        window = Window(self, "Button demo")
        window.setPosition((15, 15))
        window.setLayout(GroupLayout())

        Label(window, "Push buttons", "sans-bold")
        b = Button(window, "Plain button")

        def cb():
            print("pushed!")
        b.setCallback(cb)

        b = Button(window, "Styled", entypo.ICON_ROCKET)
        b.setBackgroundColor(Color(0, 0, 1.0, 0.1))
        b.setCallback(cb)

        Label(window, "Toggle buttons", "sans-bold")
        b = Button(window, "Toggle me")
        b.setFlags(Button.Flags.ToggleButton)

        def change_cb(state):
            print("Toggle button state: %s" % str(state))
        b.setChangeCallback(change_cb)

        Label(window, "Radio buttons", "sans-bold")
        b = Button(window, "Radio button 1")
        b.setFlags(Button.Flags.RadioButton)
        b = Button(window, "Radio button 2")
        b.setFlags(Button.Flags.RadioButton)

        Label(window, "A tool palette", "sans-bold")
        tools = Widget(window)
        tools.setLayout(BoxLayout(Orientation.Horizontal,
                                  Alignment.Middle, 0, 6))

        ToolButton(tools, entypo.ICON_CLOUD)
        ToolButton(tools, entypo.ICON_CONTROLLER_FAST_FORWARD)
        ToolButton(tools, entypo.ICON_COMPASS)
        ToolButton(tools, entypo.ICON_INSTALL)

        Label(window, "Popup buttons", "sans-bold")
        popupBtn = PopupButton(window, "Popup", entypo.ICON_EXPORT)
        popup = popupBtn.popup()
        popup.setLayout(GroupLayout())
        Label(popup, "Arbitrary widgets can be placed here")
        CheckBox(popup, "A check box")
        # popup right
        popupBtn = PopupButton(popup, "Recursive popup", entypo.ICON_FLASH)
        popupRight = popupBtn.popup()
        popupRight.setLayout(GroupLayout())
        CheckBox(popupRight, "Another check box")
        # popup left
        popupBtn = PopupButton(popup, "Recursive popup", entypo.ICON_FLASH)
        popupBtn.setSide(Popup.Side.Left)
        popupLeft = popupBtn.popup()
        popupLeft.setLayout(GroupLayout())
        CheckBox(popupLeft, "Another check box");

        window = Window(self, "Basic widgets")
        window.setPosition((200, 15))
        window.setLayout(GroupLayout())

        Label(window, "Message dialog", "sans-bold")
        tools = Widget(window)
        tools.setLayout(BoxLayout(Orientation.Horizontal,
                                  Alignment.Middle, 0, 6))

        def cb2(result):
            print("Dialog result: %i" % result)

        b = Button(tools, "Info")

        def cb():
            dlg = MessageDialog(self, MessageDialog.Type.Information, "Title",
                                "This is an information message")
            dlg.setCallback(cb2)
        b.setCallback(cb)

        b = Button(tools, "Warn")

        def cb():
            dlg = MessageDialog(self, MessageDialog.Type.Warning, "Title",
                                "This is a warning message")
            dlg.setCallback(cb2)
        b.setCallback(cb)

        b = Button(tools, "Ask")

        def cb():
            dlg = MessageDialog(self, MessageDialog.Type.Warning, "Title",
                                "This is a question message", "Yes", "No",
                                True)
            dlg.setCallback(cb2)
        b.setCallback(cb)

        import os
        import sys
        os.chdir(sys.path[0])
        try:
            icons = nanogui.loadImageDirectory(self.nvgContext(), "icons")
        except:
            try:
                icons = nanogui.loadImageDirectory(self.nvgContext(), "../icons")
            except:
                icons = nanogui.loadImageDirectory(self.nvgContext(), "../resources/icons")


        Label(window, "Image panel & scroll panel", "sans-bold")
        imagePanelBtn = PopupButton(window, "Image Panel")
        imagePanelBtn.setIcon(entypo.ICON_FOLDER)
        popup = imagePanelBtn.popup()
        vscroll = VScrollPanel(popup)
        imgPanel = ImagePanel(vscroll)
        imgPanel.setImages(icons)
        popup.setFixedSize((245, 150))

        img_window = Window(self, "Selected image")
        img_window.setPosition((710, 15))
        img_window.setLayout(GroupLayout())

        imgView = ImageView(img_window, icons[0][0])

        def cb(i):
            print("Selected item %i" % i)
            imgView.bindImage(icons[i][0])
        imgPanel.setCallback(cb)

        imgView.setGridThreshold(3)

        Label(window, "File dialog", "sans-bold")
        tools = Widget(window)
        tools.setLayout(BoxLayout(Orientation.Horizontal,
                                  Alignment.Middle, 0, 6))
        b = Button(tools, "Open")
        valid = [("png", "Portable Network Graphics"), ("txt", "Text file")]

        def cb():
            result = nanogui.file_dialog(valid, False)
            print("File dialog result = %s" % result)

        b.setCallback(cb)
        b = Button(tools, "Save")

        def cb():
            result = nanogui.file_dialog(valid, True)
            print("File dialog result = %s" % result)

        b.setCallback(cb)

        Label(window, "Combo box", "sans-bold")
        ComboBox(window, ["Combo box item 1", "Combo box item 2",
                          "Combo box item 3"])
        Label(window, "Check box", "sans-bold")

        def cb(state):
            print("Check box 1 state: %s" % state)
        chb = CheckBox(window, "Flag 1", cb)
        chb.setChecked(True)

        def cb(state):
            print("Check box 2 state: %s" % state)
        CheckBox(window, "Flag 2", cb)

        Label(window, "Progress bar", "sans-bold")
        self.progress = ProgressBar(window)

        Label(window, "Slider and text box", "sans-bold")

        panel = Widget(window)
        panel.setLayout(BoxLayout(Orientation.Horizontal,
                                  Alignment.Middle, 0, 20))

        slider = Slider(panel)
        slider.setValue(0.5)
        slider.setFixedWidth(80)

        textBox = TextBox(panel)
        textBox.setFixedSize((60, 25))
        textBox.setValue("50")
        textBox.setUnits("%")
        textBox.setFontSize(20)
        textBox.setAlignment(TextBox.Alignment.Right)

        def cb(value):
            textBox.setValue("%i" % int(value * 100))
        slider.setCallback(cb)

        def cb(value):
            print("Final slider value: %i" % int(value * 100))
        slider.setFinalCallback(cb)

        window = Window(self, "Misc. widgets")
        window.setPosition((425, 15))
        window.setLayout(GroupLayout())

        tabWidget = TabWidget(window)
        layer = tabWidget.createTab("Color Wheel")
        layer.setLayout(GroupLayout())

        Label(layer, "Color wheel widget", "sans-bold")
        ColorWheel(layer)

        layer = tabWidget.createTab("Function Graph")
        layer.setLayout(GroupLayout())
        Label(layer, "Function graph widget", "sans-bold")

        graph = Graph(layer, "Some function")
        graph.setHeader("E = 2.35e-3")
        graph.setFooter("Iteration 89")
        values = [0.5 * (0.5 * math.sin(i / 10.0) +
                         0.5 * math.cos(i / 23.0) + 1)
                  for i in range(100)]
        graph.setValues(values)
        tabWidget.setActiveTab(0)

        # Dummy tab used to represent the last tab button.
        tabWidget.createTab("+")

        def tab_cb(index):
            if index == (tabWidget.tabCount()-1):
                global counter
                # When the "+" tab has been clicked, simply add a new tab.
                tabName  = "Dynamic {0}".format(counter)
                layerDyn = tabWidget.createTab(index, tabName)
                layerDyn.setLayout(GroupLayout())
                Label(layerDyn, "Function graph widget", "sans-bold")
                graphDyn = Graph(layerDyn, "Dynamic function")

                graphDyn.setHeader("E = 2.35e-3")
                graphDyn.setFooter("Iteration {0}".format(index*counter))
                valuesDyn = [0.5 * abs((0.5 * math.sin(i / 10.0 + counter)) +
                                       (0.5 * math.cos(i / 23.0 + 1 + counter)))
                             for i in range(100)]
                graphDyn.setValues(valuesDyn)
                counter += 1
                # We must invoke perform layout from the screen instance to keep everything in order.
                # This is essential when creating tabs dynamically.
                self.performLayout()
                # Ensure that the newly added header is visible on screen
                tabWidget.ensureTabVisible(index)

        tabWidget.setCallback(tab_cb)
        tabWidget.setActiveTab(0);

        window = Window(self, "Grid of small widgets")
        window.setPosition((425, 300))
        layout = GridLayout(Orientation.Horizontal, 2,
                            Alignment.Middle, 15, 5)
        layout.setColAlignment(
            [Alignment.Maximum, Alignment.Fill])
        layout.setSpacing(0, 10)
        window.setLayout(layout)

        Label(window, "Floating point :", "sans-bold")
        floatBox = TextBox(window)
        floatBox.setEditable(True)
        floatBox.setFixedSize((100, 20))
        floatBox.setValue("50")
        floatBox.setUnits("GiB")
        floatBox.setDefaultValue("0.0")
        floatBox.setFontSize(16)
        floatBox.setFormat("[-]?[0-9]*\\.?[0-9]+")

        Label(window, "Positive integer :", "sans-bold")
        intBox = IntBox(window)
        intBox.setEditable(True)
        intBox.setFixedSize((100, 20))
        intBox.setValue(50)
        intBox.setUnits("Mhz")
        intBox.setDefaultValue("0")
        intBox.setFontSize(16)
        intBox.setFormat("[1-9][0-9]*")
        intBox.setSpinnable(True)
        intBox.setMinValue(1)
        intBox.setValueIncrement(2)

        Label(window, "Checkbox :", "sans-bold")

        cb = CheckBox(window, "Check me")
        cb.setFontSize(16)
        cb.setChecked(True)

        Label(window, "Combo box :", "sans-bold")
        cobo = ComboBox(window, ["Item 1", "Item 2", "Item 3"])
        cobo.setFontSize(16)
        cobo.setFixedSize((100, 20))

        Label(window, "Color picker :", "sans-bold");
        cp = ColorPicker(window, Color(255, 120, 0, 255));
        cp.setFixedSize((100, 20));

        def cp_final_cb(color):
            print(
                "ColorPicker Final Callback: [{0}, {1}, {2}, {3}]".format(color.r,
                                                                          color.g,
                                                                          color.b,
                                                                          color.w)
            )

        cp.setFinalCallback(cp_final_cb)

        # setup a fast callback for the color picker widget on a new window
        # for demonstrative purposes
        window = Window(self, "Color Picker Fast Callback")
        window.setPosition((425, 300))
        layout = GridLayout(Orientation.Horizontal, 2,
                            Alignment.Middle, 15, 5)
        layout.setColAlignment(
            [Alignment.Maximum, Alignment.Fill])
        layout.setSpacing(0, 10)
        window.setLayout(layout)
        window.setPosition((425, 500))
        Label(window, "Combined: ");
        b = Button(window, "ColorWheel", entypo.ICON_500PX)
        Label(window, "Red: ")
        redIntBox = IntBox(window)
        redIntBox.setEditable(False)
        Label(window, "Green: ")
        greenIntBox = IntBox(window)
        greenIntBox.setEditable(False)
        Label(window, "Blue: ")
        blueIntBox = IntBox(window)
        blueIntBox.setEditable(False)
        Label(window, "Alpha: ")
        alphaIntBox = IntBox(window)

        def cp_fast_cb(color):
            b.setBackgroundColor(color)
            b.setTextColor(color.contrastingColor())
            red = int(color.r * 255.0)
            redIntBox.setValue(red)
            green = int(color.g * 255.0)
            greenIntBox.setValue(green)
            blue = int(color.b * 255.0)
            blueIntBox.setValue(blue)
            alpha = int(color.w * 255.0)
            alphaIntBox.setValue(alpha)

        cp.setCallback(cp_fast_cb)

        self.performLayout()

        try:
            import numpy as np

            self.shader = GLShader()
            self.shader.init(
                # An identifying name
                "a_simple_shader",

                # Vertex shader
                """#version 330
                uniform mat4 modelViewProj;
                in vec3 position;
                void main() {
                    gl_Position = modelViewProj * vec4(position, 1.0);
                }""",

                """#version 330
                out vec4 color;
                uniform float intensity;
                void main() {
                    color = vec4(vec3(intensity), 1.0);
                }"""
            )

            # Draw 2 triangles
            indices = np.array(
                [[0, 2], [1, 3], [2, 0]],
                dtype=np.int32)

            positions = np.array(
                [[-1, 1, 1, -1],
                 [-1, -1, 1, 1],
                 [0, 0, 0, 0]],
                dtype=np.float32)

            self.shader.bind()
            self.shader.uploadIndices(indices)
            self.shader.uploadAttrib("position", positions)
            self.shader.setUniform("intensity", 0.5)
        except ImportError:
            self.shader = None
            pass

    def draw(self, ctx):
        self.progress.setValue(math.fmod(time.time() / 10, 1))
        super(TestApp, self).draw(ctx)

    def drawContents(self):
        if self.shader is not None:
            import numpy as np
            self.shader.bind()

            angle = time.time()

            mvp = np.array(
                [[np.cos(angle), -np.sin(angle), 0, 0],
                 [np.sin(angle), np.cos(angle), 0, 0],
                 [0, 0, 1, 0],
                 [0, 0, 0, 1]],
                dtype=np.float32
            )

            mvp[0:3, 0:3] *= 0.25
            mvp[0, :] *= self.size()[1] / self.size()[0]

            self.shader.setUniform("modelViewProj", mvp)
            self.shader.drawIndexed(gl.TRIANGLES, 0, 2)
        super(TestApp, self).drawContents()

    def keyboardEvent(self, key, scancode, action, modifiers):
        if super(TestApp, self).keyboardEvent(key, scancode,
                                              action, modifiers):
            return True
        if key == glfw.KEY_ESCAPE and action == glfw.PRESS:
            self.setVisible(False)
            return True
        return False

if __name__ == "__main__":
    nanogui.init()
    test = TestApp()
    test.drawAll()
    test.setVisible(True)
    nanogui.mainloop()
    del test
    gc.collect()
    nanogui.shutdown()



#!/usr/bin/env python3

# import sys
import pymrpt
import argparse

# args
parser = argparse.ArgumentParser()
parser.add_argument('rawlog', help='Rawlog file.')
parser.add_argument('-c', '--config', help='Config file.')
parser.add_argument('-o', '--output', help='Save occupancy grid as .gridmap file.')
parser.add_argument('-i', '--image', help='Save occupancy grid as image (.png, .bmp) file.')
parser.add_argument('-r', '--resolution', help='Window resolution. Default: 800x600')
args = parser.parse_args()

# get filenames from args
rawlog_filename = args.rawlog
config_filename = args.config
output_filename = args.output
image_filename = args.image

# load .rawlog
rawlog_file = pymrpt.utils.CFileGZInputStream(rawlog_filename)

# map builder
options = pymrpt.slam.CMetricMapBuilderRBPF.TConstructionOptions()
if config_filename:
    options.loadFromConfigFileName(config_filename, 'MappingApplication')
options.dumpToConsole()
map_builder = pymrpt.slam.CMetricMapBuilderRBPF(options)

# get window resolution
if args.resolution:
    resolution_str = args.resolution
else:
    resolution_str = '800x600'

# gui
try:
    win3D = pymrpt.gui.CDisplayWindow3D("rbpf_slam", int(resolution_str.split('x')[0]), int(resolution_str.split('x')[1]))
except:
    win3D = pymrpt.gui.CDisplayWindow3D("rbpf_slam", 800, 600)

# loop
entry = 0
while True:
    # get action observation pair
    next_entry, act, obs, entry = pymrpt.obs.CRawlog.readActionObservationPair(
        rawlog_file, entry)
    if not next_entry:
        break
    else:
        entry += 1
    print(('Processing entry: {}.'.format(entry)))

    # process
    map_builder.processActionObservation(act, obs)

    # get most likely map
    metric_map = map_builder.mapPDF.getCurrentMostLikelyMetricMap()

    # get pose estimation
    pose_est = map_builder.getCurrentPoseEstimation()
    particles_count = pose_est.ctx().particlesCount()

    # get particles paths
    lines_object = pymrpt.opengl.CSetOfLines.Create()
    for i in range(particles_count):
        path = map_builder.mapPDF.getPath(i)
        x = 0.0
        y = 0.0
        z = 0.0
        for p in path:
            lines_object.ctx().appendLine(x, y, z + 0.001, p.x, p.y, p.z + 0.001)
            x = float(p.x)
            y = float(p.y)
            z = float(p.z)
    lines_object.ctx().setColor(pymrpt.utils.TColorf(1., 0., 0.))

    # get ellipsoids at time t
    dummy_path = map_builder.mapPDF.getPath(0)
    k_s = list(range(len(dummy_path)))
    k_s.reverse()
    ellipsoid_objects = []
    for k in k_s:
        pose_particles = map_builder.mapPDF.getEstimatedPosePDFAtTime(k)
        ellipsoid_object = pymrpt.opengl.CEllipsoid2D.Create()
        ellipsoid_object.ctx().setFromPosePDF(pose_particles)
        ellipsoid_object.ctx().setColor(pymrpt.utils.TColorf(0., 1., 0.))
        ellipsoid_objects.append(ellipsoid_object)

    # update gui
    map_object = metric_map.getAs3DObject()

    scene = win3D.get3DSceneAndLock()
    scene.ctx().clear()
    scene.ctx().insert(map_object)
    scene.ctx().insert(lines_object)
    for ellipsoid_object in ellipsoid_objects:
        scene.ctx().insert(ellipsoid_object)
    win3D.unlockAccess3DScene()
    win3D.forceRepaint()

# get last built map and save it as occupancy grid to file
simple_map = map_builder.getCurrentlyBuiltMap()
occ_map = pymrpt.maps.COccupancyGridMap2D()
occ_map.loadFromSimpleMap(simple_map)
if output_filename:
    occ_map.saveAsBitmapFile(output_filename)
    print(('Saved map as {}'.format(output_filename)))
if image_filename:
    occ_map.saveAsBitmapFile(image_filename)
    print(('Saved map image as {}'.format(output_filename)))
print('Done.')
input('Press key to quit.')



#!/usr/bin/env python3

import os
import sys
import pymrpt
import argparse
from time import sleep

# args
parser = argparse.ArgumentParser()
parser.add_argument('config', help='Config file.')
parser.add_argument('-d', '--delay', help='Time delay in seconds. Default: 0.2')
parser.add_argument('-r', '--resolution', help='Window resolution. Default: 800x600')
args = parser.parse_args()

# get filenames from args
config_filename = args.config

# get configuration
if not os.path.exists(config_filename):
    print('Error. Config file not found.')
    print('Quit.')
    sys.exit(1)

config_file = pymrpt.utils.CConfigFile(config_filename)
print('Load config file {}.'.format(config_filename))
sec_name = 'LocalizationExperiment'

rawlog_filename = config_file.read_string(sec_name, "rawlog_file", "")
map_filename = config_file.read_string(sec_name, "map_file", "")


# default filenames are relative so we need to change our dir to supplied config file dir
curr_dir = os.path.abspath(os.curdir)
config_dir = os.path.dirname(os.path.abspath(config_filename))
# rawlog
if not os.path.exists(os.path.abspath(rawlog_filename)):
    os.chdir(config_dir)
    if not os.path.exists(os.path.abspath(rawlog_filename)):
        print('Error. Rawlog file not found.')
        print('Quit.')
        sys.exit(1)
    else:
        rawlog_filename = os.path.abspath(rawlog_filename)
        os.chdir(curr_dir)
rawlog_file = pymrpt.utils.CFileGZInputStream(rawlog_filename)
print('Load rawlog file {}.'.format(rawlog_filename))

# map
if not os.path.exists(os.path.abspath(map_filename)):
    os.chdir(config_dir)
    if not os.path.exists(os.path.abspath(map_filename)):
        print('Error. Map file not found.')
        print('Quit.')
        sys.exit(1)
    else:
        map_filename = os.path.abspath(map_filename)
        os.chdir(curr_dir)

# options
# kld
pdf_prediction_options = pymrpt.slam.TMonteCarloLocalizationParams()
pdf_prediction_options.KLD_params.loadFromConfigFileName(config_filename, 'KLD_options')
pdf_prediction_options.KLD_params.dumpToConsole()
# pf
pf_options = pymrpt.bayes.CParticleFilter.TParticleFilterOptions()
pf_options.loadFromConfigFileName(config_filename, 'PF_options')
pf_options.dumpToConsole()
# maps
map_list = pymrpt.maps.TSetOfMetricMapInitializers()
map_list.loadFromConfigFileName(config_filename, 'MetricMap')
map_list.dumpToConsole()

## setup
metric_map = pymrpt.maps.CMultiMetricMap()
metric_map.setListOfMaps(map_list)


# load map
if (map_filename.endswith('.simplemap')
        or map_filename.endswith('.simplemap.gz')):
    simple_map = pymrpt.maps.CSimpleMap.Create()
    map_file = pymrpt.utils.CFileGZInputStream(map_filename)
    map_file.ReadObject(simple_map)
    metric_map.loadFromProbabilisticPosesAndObservations(simple_map.ctx())
elif (map_filename.endswith('.gridmap')
        or map_filename.endswith('.gridmap.gz')):
    occ_map = pymrpt.maps.COccupancyGridMap2D.Create()
    map_file = pymrpt.utils.CFileGZInputStream(map_filename)
    map_file.ReadObject(occ_map)
    # overwrite gridmap
    for i in range(len(metric_map.maps)):
        if metric_map.maps[i].ctx().GetRuntimeClass().className == 'COccupancyGridMap2D':
            metric_map.maps[i] = occ_map
else:
    print('Error. Can not load map from unknown extension.')
    print('Quit.')
    sys.exit(1)
print('Load map file {}.'.format(map_filename))


# get window resolution
if args.resolution: resolution_str = args.resolution
else: resolution_str = '800x600'

# gui
try:
    win3D = pymrpt.gui.CDisplayWindow3D("pf_localization", int(resolution_str.split('x')[0]), int(resolution_str.split('x')[1]))
except:
    win3D = pymrpt.gui.CDisplayWindow3D("pf_localization", 800, 600)

# initial scene
map_object = metric_map.maps[0].ctx().getAs3DObject()

scene_ptr = win3D.get3DSceneAndLock()
scene_ptr.ctx().clear()
scene_ptr.ctx().insert(map_object)
win3D.unlockAccess3DScene()
win3D.forceRepaint()

# mcl
pdf = pymrpt.slam.CMonteCarloLocalization2D()
pdf.options = pdf_prediction_options
pdf.options.metricMap = metric_map

pf = pymrpt.bayes.CParticleFilter()
pf.m_options = pf_options

# initialize pdf
pdf.resetUniformFreeSpace(metric_map.maps[0].pointer(), 0.7, 40000)

# loop
entry = 0
while True:
    # get action observation pair
    next_entry, act, obs, entry = pymrpt.obs.CRawlog.readActionObservationPair(rawlog_file, entry)
    if not next_entry: break
    else: entry += 1
    print('Processing entry: {}.'.format(entry))

    # get covariance and mean
    cov, mean = pdf.getCovarianceAndMean()

    # get particles
    particles_object = pdf.getAs3DObject()

    # get laserscan
    points_map = obs.buildAuxPointsMap()
    laserscan_object = points_map.getAs3DObject()
    laserscan_object.ctx().setPose(pymrpt.poses.CPose3D(mean))
    laserscan_object.ctx().setColor(pymrpt.utils.TColorf(1.,0.,0.))

    # update pf
    act_ptr = pymrpt.obs.CActionCollection.Create()
    act_ptr.ctx(act)
    obs_ptr = pymrpt.obs.CSensoryFrame.Create()
    obs_ptr.ctx(obs)
    stats = pf.executeOn(pdf, act_ptr, obs_ptr)

    # update scene
    scene_ptr = win3D.get3DSceneAndLock()
    scene_ptr.ctx().clear()
    scene_ptr.ctx().insert(map_object)
    scene_ptr.ctx().insert(particles_object)
    scene_ptr.ctx().insert(laserscan_object)
    win3D.unlockAccess3DScene()
    win3D.forceRepaint()

    # sleep
    if args.delay:
        try:
            sleep(float(args.delay))
        except:
            sleep(0.2)
    else: sleep(0.2)

print()
print('Done.')
input('Press key to quit.')



#!/usr/bin/env python3

"""
Current script replaces instances of multiple namespaces with the modern
C++17-equivalent nested namespaces

.. seealso::
* `<https://github.com/MRPT/mrpt/issues/752>_`
* `<http://www.open-std.org/jtc1/sc22/wg21/docs/papers/2014/n4230.html>_`

"""

import click
import sh
import re
from typing import List


@click.command()
@click.option('-r', '--root-dir', help='Top-level directory to operate in',
              required=True, type=click.Path(exists=True))
@click.option('-e', '--extensions', help='Extensions which are to be considered', multiple=True)
def main(root_dir: str, extensions: List[str]):
    # list of files to operate in
    fpaths = []  # type: List[str]
    # files that need to be checked manually
    files_manual_check = []  # type: List[str]
    num_files_modified = 0  # type: int

    # find all the files we're interested in
    for e in extensions:
        fpaths.extend([i.rstrip() for i in sh.find(root_dir, "-iname",
                                                   "*.{}".format(e))])

    expr_text = "namespace (\w*)[\s]{"
    expr = re.compile(expr_text)
    # we now have a list of files
    for fpath in fpaths:
        with open(fpath, 'rU') as f:
            old_conts = f.read()
        # do we have more than a single match
        ns_expr_results = expr.findall(old_conts)
        times_found = len(ns_expr_results)
        if times_found > 1:
            start_str = "\s*".join([expr_text] * times_found)
            # match whitespace characters or comments after '}'
            ending_str_after = "(\s*)(//.*)?(\s*)"
            end_str = ending_str_after.join(['}'] * times_found)
            end_str += ending_str_after

            valid_start_results = re.search(start_str, old_conts)
            valid_end_results = re.search(end_str, old_conts)

            if not valid_start_results or not valid_end_results:
                files_manual_check.append(fpath)
                continue
            # compose new contents
            new_conts = old_conts

            # convert the start
            # grab the namespace names
            ns_names = [g for g in valid_start_results.groups()]
            # compose the new start
            new_start = " ".join(["namespace", "::".join(ns_names), ])
            new_start = "\n".join([new_start, "{"])

            new_conts = "".join([new_conts[:valid_start_results.start()],
                                 new_start,
                                 new_conts[valid_start_results.end():],
                                 "\n"])

            # need to update the end characters position
            valid_end_results = list(re.finditer(end_str, new_conts))[-1]
            # convert the end
            # convert only the count of instances that you are
            # interested in, starting from the bottom
            new_end = "}"
            new_conts = "".join([new_conts[:valid_end_results.start()],
                                 new_end,
                                 "\n",
                                 new_conts[valid_end_results.end():],
                                 "\n"])

            # flush the new contents
            with open(fpath, 'w') as f:
                num_files_modified += 1
                f.write(new_conts)


    print("{} files modified".format(num_files_modified))
    print("{} files to check manually:\n{}".format(
        len(files_manual_check),
        '\n'.join(["\t{}".format(i) for i in
                   files_manual_check])))

if __name__ == "__main__":
    main()



# python/example_icons.py -- Python version of an example application that shows
# all available Entypo icons as they would appear in NanoGUI itself.  For a C++
# implementation, see '../src/example_icons.cpp'.
#
# NanoGUI was developed by Wenzel Jakob <wenzel.jakob@epfl.ch>.
# The widget drawing code is based on the NanoVG demo application
# by Mikko Mononen.
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE.txt file.

# Developer note: need to make a change to this file?
# Please raise an Issue on GitHub describing what needs to change.  This file
# was generated, so the scripts that generated it needs to update as well.

import gc

import nanogui
from nanogui import Screen, Window, Widget, GridLayout, VScrollPanel, Button
from nanogui import entypo

if __name__ == "__main__":
    nanogui.init()

    width      = 1000
    half_width = width // 2
    height     = 800

    # create a fixed size screen with one window
    screen = Screen((width, height), "NanoGUI Icons", False)
    window = Window(screen, "All Icons")
    window.setPosition((0, 0))
    window.setFixedSize((width, height))

    # attach a vertical scroll panel
    vscroll = VScrollPanel(window)
    vscroll.setFixedSize((width, height))

    # vscroll should only have *ONE* child. this is what `wrapper` is for
    wrapper = Widget(vscroll)
    wrapper.setFixedSize((width, height))
    wrapper.setLayout(GridLayout())  # defaults: 2 columns

    # NOTE: don't __dict__ crawl in real code!
    # this is just because it's more convenient to do this for enumerating all
    # of the icons -- see cpp example for alternative...
    for key in entypo.__dict__.keys():
        if key.startswith("ICON_"):
            b = Button(wrapper, "entypo.{0}".format(key), entypo.__dict__[key])
            b.setIconPosition(Button.IconPosition.Left)
            b.setFixedWidth(half_width)

    screen.performLayout()
    screen.drawAll()
    screen.setVisible(True)

    nanogui.mainloop()

    del screen
    gc.collect()

    nanogui.shutdown()
